"""
This file demonstrates writing tests using the unittest module. These will pass
when you run "manage.py test".

Replace this with more appropriate tests for your application.
"""

from django.test import TestCase

import models

class MailingListTreeTestCase(TestCase):
    '''Test mailing lists member resolution in a classical tree setup.'''
    def setUp(self):
        self.users = []
        self.mls = []
        for i in range(20):
            self.users.append(models.User.objects.create(username='%s' % i))
        sublist = []
        for i in range(19, -1, -1):
            self.mls.insert(0, models.MailingList.objects.create(
                    name='%s' % i))
            self.mls[0].members = [self.users[i]]
            self.mls[0].mailing_list_members = sublist
            sublist = [ self.mls[0] ]

    def test_mailing_list_recursive(self):
        """
        Tests that 1 + 1 always equals 2.
        """
        for i in range(19, -1, -1):
            members = self.mls[i].recursive_members()
            self.assertEqual(len(members), 20-i)

    def test_mailing_list_recursive_with_origin(self):
        """
        Tests that 1 + 1 always equals 2.
        """
        for i in range(19, -1, -1):
            mwo = self.mls[i].recursive_members_with_origin()
            self.assertEqual(len(mwo), 20-i)
            for j in range(i, 20):
                self.assertEqual(mwo[self.users[j]], set(self.mls[i:j+1]))


class MailingListCycle(TestCase):
    '''Test mailing lists member resolution in a cyclic setup.'''
    def setUp(self):
        self.users = []
        self.mls = []
        for i in range(20):
            self.users.append(models.User.objects.create(username='%s' % i))
        sublist = []
        for i in range(19, -1, -1):
            self.mls.insert(0, models.MailingList.objects.create(
                    name='%s' % i))
            self.mls[0].members = [self.users[i]]
            self.mls[0].mailing_list_members = sublist
            sublist = [ self.mls[0] ]
        self.mls[19].mailing_list_members = [ self.mls[0] ]

    def test_mailing_list_recursive(self):
        """
        Tests that 1 + 1 always equals 2.
        """
        for i in range(19, -1, -1):
            members = self.mls[i].recursive_members()
            self.assertEqual(len(members), 20)

    def test_mailing_list_recursive_with_origin(self):
        """
        Tests that 1 + 1 always equals 2.
        """
        for i in range(19, -1, -1):
            mwo = self.mls[i].recursive_members_with_origin()
            self.assertEqual(len(mwo), 20)
            for j in range(i, 20):
                self.assertEqual(mwo[self.users[j]], set(self.mls))


class BaseTestCase(TestCase):
        def setUp(self):
            from django.contrib.auth.models import User
            from django.core.files.base import ContentFile
            from .models import (DocbowProfile, Document, FileType,
                    AttachedFile)
            import tempfile

            self.users = []
            for i in range(10):
                self.users.append(
                        User.objects.create(username='user-%s' % i, 
                            email='user-%s@example.com' % i))
                DocbowProfile.objects.create(user=self.users[-1],
                    personal_email='personal-email-user-%s@example.com' % i)
            self.filetypes = []
            for i in range(10):
                self.filetypes.append(
                    FileType.objects.create(name='filetype-%s' % i))
            self.documents = []
            for i in range(10):
                self.documents.append(
                    Document.objects.create(sender=self.users[(i+2) % 10],
                        filetype=self.filetypes[i]))
                self.documents[-1].to_user = [self.users[i % 10], 
                        self.users[(i+1) % 10]]
                for j in range(2):
                    attached_file = AttachedFile(name='file%s' % j,
                            document=self.documents[-1],
                            kind=None)
                    attached_file.content.save('file%s.pdf' % j, ContentFile('coucou'))
                    attached_file.save()
                self.documents[-1].post()

class BasicTestCase(BaseTestCase):
    def test_notification_mail(self):
        from django.core import mail
        from django.core import management
        import re

        management.call_command('notify')
        self.assertEquals(len(mail.outbox), 20)
        outbox = sorted(mail.outbox, key=lambda m: tuple(sorted(m.to)))
        MAIL_LINK_RE = re.compile('https?://[^/]+/inbox/\d+/')
        for message, i in zip(outbox, range(20)):
            recipient = self.users[i / 2]
            emails = [recipient.docbowprofile.personal_email, recipient.email]
            self.assertEquals(set(message.to), set(emails))
            match = MAIL_LINK_RE.search(message.body)
            self.assertIsNotNone(match)

class UtilsTestCase(BaseTestCase):
        def setUp(self):
            super(UtilsTestCase, self).setUp()
            self.user1, self.user2 = self.users[:2]
            self.user2.docbowprofile.is_guest = True
            self.user2.docbowprofile.save()

        def test_generate_filename(self):
            from . import models
            self.assertNotEquals(models.generate_filename(None, 'xxx.pdf'),
                models.generate_filename(None, 'xxx.pdf'))

        def test_all_emails(self):
            from . import models
            self.assertEquals(set(models.all_emails(self.user1)),
                set((self.user1.email, self.user1.docbowprofile.personal_email)))

        def test_document_manager(self):
            from . import models
            with self.assertNumQueries(2):
                list(models.Document.objects.all())

        def test_document_accessors(self):
            document = self.documents[0]
            self.assertEqual(document.filenames(),
                    'file0.pdf, file1.pdf')
            self.assertEqual(set(document.user_human_to()),
                    set(['user-0', 'user-1']))
            self.assertEqual(document.group_human_to(), [])
            self.assertEqual(set(document.human_to()),
                    set(['user-0', 'user-1']))
            self.assertIsNotNone(document.filename_links())
            self.assertEquals(set(document.to()), set((self.user1, self.user2)))
            self.assertEquals(dict(document.to_with_origin()),
                    {self.user1: set([ '--direct--' ]), self.user2: set(['--direct--'])})
            self.assertEqual(sorted(map(lambda x: x.pk, document.delivered_to())),
                    sorted([ self.user1.pk, self.user2.pk]))

        def test_is_guest(self):
            from . import models
            self.assertTrue(models.is_guest(self.user2))

class DelegatesTestCase(BaseTestCase):
        def setUp(self):
            super(DelegatesTestCase, self).setUp()
            from django.contrib.auth.models import User
            from .models import Delegation
            delegate = User(username='delegate')
            delegate.set_password('delegate')
            delegate.save()
            Delegation.objects.create(by=self.users[0], to=delegate)
            Delegation.objects.create(by=self.users[1], to=delegate)

        def test_inbox_by_document(self):
            from django.test import Client
            c = Client()
            c.login(username='delegate', password='delegate')
            response = c.get('/inbox_by_document/%s/' % self.documents[0].id)
            self.assertEqual(response.status_code, 302)
            # Mailbox default ordering is -date so the last recipient mailbox
            # is the first mailbox to be returned
            self.assertTrue(response['Location'].endswith('/inbox/%s/' % self.documents[0].id), response['Location'])
